"""
AbstractDescriptor provides the base class for descriptors in the
'worktoy.attr' module. It inherits the descriptor protocol from the
'worktoy.static.AbstractObject' class and enhances it. Please note that
instances of this class and subclasses are intended to be defined inside a
class body. The class created from this body 'owns' the descriptor instance.

The enhancements include:

# Instance specific accessor functions.

This means that an instance owned by a class can have any callable
assigned to each accessor function. The type signatures expected for these
are as follows:
# #  getter(instance: Any, ) -> Any:
# #  setter(instance: Any, value: Any) -> None:
# #  deleter(instance: Any) -> None:

This allows for two different paths for descriptors that diverge as they
subclass AbstractDescriptor:
#  1. The owning class must explicitly specify each accessor function for
example by use of decorators (foreshadowing). (This is similar to the
builtin 'property' class).

#  2. The descriptor class itself implements each accessor function. This
will always be less flexible than the first option, but can reuse advanced
behaviour across multiple descriptors and even classes. A common use-case
is lazy instantiation of a wrapped class such that the wrapped class is
not instantiated until requested by a '__get__' call.

#  Accessor hooks.
The accessors have hooks in their operation flow. The 'getter' has pre,
fix and post hooks invoked before, during and after the value returns.
'setter' and 'deleter' have pre- and post-hooks invoked before and after
respectively.

#  Use Cases
The hooks surrounding each accessor provide a convenient place to
implement logging, debugging and validation functionality. Other common
use cases are specific to the accessor such as caching. The hooks also
allow for more imaginative behaviour even to the point of being
unconventional. The less conventional the behaviour, the more thorough the
documentation.

#  Hooks
#  #  '__get__'
The 'getter' is a general concept in computer science, requiring the
following two steps:
1.  Retrieval of the object to be returned.
2.  Returning the retrieved object.
The 'setter' and 'deleter' in contrast are single step operations. This
extra step required by the 'getter' is accounted for by the third hook,
'fix' invoked between the retrieval and return steps.

#  #  #  preGET(instance: Any, value: Any=None) -> Any:
Each hook is invoked in order registered. The first hook receives no
value, the remaining hooks receive the value returned by the previous
hook. If the final hook returns a value other than 'None', the normal
retrieval operation is skipped and proceeds with the value.

Please note, that a pre hook returning 'None' after an earlier hook has
returned a value, it is not replaced. If for some reason a pre hook wishes
to cancel a previously set value, it must raise 'BadCache'. When the flow
catches 'BadCache', it ignores values returned by pre hooks and retrieves
the value normally. This is highly discouraged, if a value does need some
adjustment, use the fix hook defined below:

#  #  #  Retrieval of Value
The 'setter' and 'deleter' hooks receive the existing value or the error
encountered when attempting retrieval. By setting keyword argument
'_root', the 'getter' hooks are entirely ignored. To facilitate this,
the underlying retrieval operation must raise 'MissingVariable' which is
the only exception caught in this flow. The alternative of just catching
and returning any 'Exception' object is poor practice. By requiring that
the retrieval procedure is able to recognize an allowable exception,
such as when a value has not been set for the first time, the flow is
safer and cleaner.

Final word: It is highly discouraged to implement a retrieval operation
that deliberately returns 'None' for any reason. Doing so invites
confusion whether the value is set. In particular, the following two
conditions should never both be true:

if getattr(instance, key, None) is None and hasattr(instance, key):
  raise ...  # Bad, awful, terrible!

#  #  #  FIX(instance: Any, value: Any) -> Any:
The hook in between retrieval and return is the 'fix' hook. This hook
allows modification of the value retrieved before it is returned. When
multiple hooks are registered, the value returned by the previous hook is
passed to the next hook. If a hook returns 'None', the previous value is
used. If a hook wants to cancel a previously set value, it must raise
'BadCache'.

#  #  #  Return Value
Following the fix hooks, the value is returned, but it happens in a
try/finally block to allow for post hooks after returning the value. A
common misconception is that nothing further can happen after 'return',
but finally clauses are executed after the return statement. Please note
that this limits what should be done in the finally block. Specifically,
the finally block can raise exceptions where appropriate, which replaces
any waiting exceptions, but this is allowed. What is bad practice is for
the 'finally' clause to: break, continue, yield or return. Beginning in
Python 3.14 doing so triggers a 'SyntaxWarning' (same in the alpha builds
of 3.15 available at the time of writing).

#  #  #  postGET(instance: Any, value: Any) -> None:
After the value has been returned, the post hooks are invoked. These are
not able to change the value returned, but any further processing
involving the returned value is waiting for the finally block. For example:

#  Test script
from __future__ import annotations

import sys

def foo() -> int:
  try:
    return 69
  finally:
    print('Finally block!')

class Num:
  def __get__(self, instance: Any, owner: type) -> Any:
    if instance is None:
      return self
    return getattr(instance, '__test_number__', 0)

  def __set__(self, instance: Any, value: Any) -> None:
    setattr(instance, '__test_number__', value)

class Bar:
  __test_number__ = 420
  num = Num()
  def __init__(self, *args) -> None:
    print('initiating Bar object!')
    for arg in args:
      if isinstance(arg, int):
        self.__test_number__ = arg
        break

  def __setattr__(self, key: str, value: Any) -> None:
    print('Setting %s to %s' % (key, value))
    return object.__setattr__(self, key, value)

  def __str__(self) -> str:
    return 'Bar[num=%d]' % self.num

def main() -> int:
  bar = Bar()
  bar.num = foo()
  print(bar)
  return 0

if __name__ == '__main__':
  sys.exit(main())

The above script receives the following stdout:

  initiating Bar object!
  Finally block!  # The finally block in foo()
  Setting num to 69  #  Public set call
  # The public set call had to wait for the finally block in foo()
  Setting __test_number__ to 69  # Private set call
  Bar[num=69]

In summary, 'ham = eggs()' still waits for the finally block in eggs()
even if it has already returned a value. So the finally block does provide
a place for a 'post' hook, just not post by very much. This leaves the
post hook as more of an academic exercise, as it is quite literally
impossible for anything to have happened in between the return statement
and the 'finally' clause.

#  preGET(instance: Any, value: Any) -> Any:
Next, the value has been retrieved from the normal procedure or is a
cached value. This value is then passed to the fix hooks. These hooks can
further augment the value, or validate it.

#  fixGET(instance: Any, value: Any) -> Any:
Following the fix hooks, the value is returned from '__get__' in a
try/finally with the finally block

#  set
Before invoking the hook, the existing value is retrieved and passed to
the pre hooks along with the new value:
#  preSET(instance: Any, oldValue: Any, newValue: Any) -> None:
After the pre hooks, the new value is set and the post hooks are invoked:
#  postSET(instance: Any, oldValue: Any, newValue: Any) -> None:



Because
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any



"""AccessorHook hooks into the namespace system and logs each call to
__getitem__ and __setitem__, but does not otherwise modify any behaviour. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ...waitaminute import MissingVariable
from . import AbstractHook

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Callable, Self
  from worktoy.mcls import AbstractNamespace as ASpace


class _Field:
  """Key is a descriptor that logs calls to __setitem__ on the
  namespace object. """

  __field_name__ = None
  __field_owner__ = None
  __private_name__ = None

  def __init__(self, *args) -> None:
    """Initialize the Key object."""
    if args:
      self.__private_name__ = args[0]

  def getFallbackValue(self) -> Any:
    """Get the fallback value."""

  def __set_name__(self, owner: type, name: str) -> None:
    """Set the name of the field and the owner of the field."""
    self.__field_name__ = name
    self.__field_owner__ = owner

  def _getFieldName(self, ) -> str:
    """Getter-function for the field name."""
    if self.__field_name__ is None:
      raise MissingVariable('__field_name__', str)
    return self.__field_name__

  def _getFieldOwner(self, ) -> type:
    """Getter-function for the field owner."""
    if self.__field_owner__ is None:
      raise MissingVariable('__field_owner__', type)
    return self.__field_owner__

  def _getPrivateName(self) -> str:
    """Getter-function for the private name."""
    fieldName = self._getFieldName()
    if self.__private_name__ is None:
      if fieldName.startswith('__') and fieldName.endswith('__'):
        return '_pvt_%s' % fieldName[2:-2]
      words = []
      word = []
      for char in fieldName:
        if char.isupper():
          if word:
            words.append(''.join(word))
            word = []
        word.append(char.lower())
      return '__%s__' % '_'.join(words)
    return self.__private_name__

  def __get__(self, instance: object, owner: type) -> Any:
    """Get the name of the function."""
    if instance is None:
      return self
    pvtName = self._getPrivateName()
    return getattr(instance, pvtName, self.getFallbackValue())


class _Key(_Field):
  """Key is a descriptor that logs calls to __getitem__ on the
  namespace object. """
  pass


class _Value(_Field):
  """Value is a descriptor that logs calls to __setitem__ on the
  namespace object. """
  pass


class _GetItem:
  """Represents a __getitem__ call. """

  __pvt_key__ = None
  __pvt_val__ = None
  __pvt_err__ = None

  key = _Key('__pvt_key__')
  value = _Value('__pvt_val__')
  err = _Value('__pvt_err__')

  def __init__(self, key: str, value: Any, ) -> None:
    """Initialize the __getitem__ call."""
    self.__pvt_key__ = key
    if isinstance(value, Exception):
      self.__pvt_err__ = value
    elif value is not None:
      self.__pvt_val__ = value


class _SetItem:
  """Represents a __setitem__ call. """

  __pvt_key__ = None
  __new_val__ = None
  __old_val__ = None

  key = _Key('__pvt_key__')
  new_val = _Value('__new_val__')
  old_val = _Value('__old_val__')

  def __init__(self, *args) -> None:
    _key, _new, _old = [*args, None, None, None][:3]
    if _key is not None:
      self.__pvt_key__ = _key
    if _new is not None:
      self.__new_val__ = _new
    if _old is not None:
      self.__old_val__ = _old


class AccessorHook(AbstractHook):
  """AccessorHook hooks into the namespace system and logs each call to
  __getitem__ and __setitem__, but does not otherwise modify any
  behaviour. """

  __log_name__ = '__accessor_lines__'

  def _getLogName(self, ) -> str:
    """Getter-function for log name."""
    if self.__log_name__ is None:
      raise MissingVariable('__log_name__', str)
    return self.__log_name__

  def getItemHook(self, space: ASpace, key: str, value: Any) -> bool:
    """Hook for getItem. This is called before the __getitem__ method of
    the namespace object is called. The default implementation does nothing
    and returns False. """
    pvtName = self._getLogName()
    item = _GetItem(key, value)
    instance = self.getOwningNamespace()
    setattr(instance, pvtName, item)
    return False

  def setItemHook(
      self,
      space: ASpace,
      key: str,
      value: object,
      oldValue: object
  ) -> bool:
    """Hook for setItem. This is called before the __setitem__ method of
    the namespace object is called. The default implementation does nothing
    and returns False. """
    pvtName = self._getLogName()
    item = _SetItem(key, value, oldValue)
    instance = self.getOwningNamespace()
    setattr(instance, pvtName, item)
    return False



"""
TestHistDict tests the HistDict class.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from unittest import TestCase

from worktoy.static import HistDict, ItemCall

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Optional, Union, Self, Callable, TypeAlias, Never


class TestHistDict(TestCase):
  """
  TestHistDict tests the HistDict class.
  """

  def setUp(self, ) -> None:
    """
    Set up the test case.
    """
    super().setUp()



"""ParseException is a custom exception raised to indicate that a keyword
argument parsing scheme has failed. Callers should handle this exception
for example by moving on to another parsing scheme."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import _Attribute
from ..text import monoSpace

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Optional, Self, TypeAlias


class ParseException(ValueError):
  """ParseException is a custom exception raised to indicate that a keyword
  argument parsing scheme has failed. Callers should handle this exception
  for example by moving on to another parsing scheme."""

  name = _Attribute()
  type_ = _Attribute()

  def __init__(self, name: str, type_: type) -> None:
    """Initialize the ParseException object."""
    self.name = name
    self.type_ = type_
    infoSpec = """Unable to parse argument '%s' of type '%s' whilst 
    parsing keyword arguments!"""
    info = monoSpace(infoSpec % (name, type_.__name__))
    ValueError.__init__(self, info)

  def __eq__(self, other: Any) -> bool:
    """Compare the ParseException object with another object."""
    cls = type(self)
    if isinstance(other, cls):
      if self.name == other.name:
        if self.type_ == other.type_:
          return True
    return False



"""AbstractNamespace class provides a base class for custom namespace
objects used in custom metaclasses."""
#  AGPL-3.0 license
#  Copyright (c) 2024-2025 Asger Jon Vistisen
from __future__ import annotations

from ..static import HistDict
from ..text import typeMsg, monoSpace
from ..waitaminute import MissingVariable, HookException
from ..waitaminute import HashError, TypeException
from . import Base, Types, Spaces
from .hooks import AbstractHook

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Self, TypeAlias

  Bases: TypeAlias = tuple[type, ...]


class AbstractNamespace(HistDict):
  """AbstractNamespace class provides a base class for custom namespace
  objects used in custom metaclasses."""

  #  Reserved private names
  __meta_class__ = None
  __class_name__ = None
  __base_classes__ = None
  __key_args__ = None
  __owner_hooks_list_name__ = '__hook_objects__'
  __hash_value__ = None

  def getClassBases(self) -> Bases:
    """Returns the base classes of the class under creation."""
    return (*self.__base_classes__,)

  def _computeHash(self, ) -> int:
    """
    Computes the hash value for the future class.
    """
    clsName = self.getClassName()
    baseNames = [base.__name__ for base in self.getClassBases()]
    mclsName = self.getMetaclass().__name__
    return hash((clsName, *baseNames, mclsName,))

  def getHash(self, **kwargs) -> int:
    """
    Resolves the hash value for the future class.
    """
    computedHash = self._computeHash()
    if self.__hash_value__ is None:
      if kwargs.get('_recursion', False):
        raise RecursionError
      self.__hash_value__ = computedHash
      return self.getHash(_recursion=True)
    if not isinstance(self.__hash_value__, int):
      raise TypeException('__hash_value__', self.__hash_value__, int)
    if self.__hash_value__ != computedHash:
      raise HashError(self.__hash_value__, computedHash, )
    return self.__hash_value__

  @classmethod
  def getHookListName(cls, ) -> str:
    """Getter-function for the name of the hook list. """
    if TYPE_CHECKING:
      assert isinstance(cls, dict)
    return cls.__owner_hooks_list_name__

  def getHooks(self, owner: type = None) -> list[AbstractHook]:
    """Getter-function for the AbstractHook classes. """
    pvtName = self.getHookListName()
    cls = type(self)
    hooks = getattr(cls, pvtName, [])
    out = []
    for hook in hooks:
      out.append(hook.__get__(self, cls))
    return out

  @classmethod
  def addHook(cls, hook: AbstractHook) -> None:
    """Adds a hook to the list of hooks. """
    if TYPE_CHECKING:
      assert isinstance(cls, dict)
    if not isinstance(hook, AbstractHook):
      raise TypeError(typeMsg('hook', hook, AbstractHook))
    existingHooks = getattr(cls, cls.getHookListName(), [])
    setattr(cls, cls.getHookListName(), [*existingHooks, hook])

  def __init__(self, mcls: type, name: str, bases: Base, **kwargs) -> None:
    self.__meta_class__ = mcls
    self.__class_name__ = name
    if len(bases) > 1:
      e = """The 'worktoy' package does not support multiple inheritance!"""
      raise TypeError(monoSpace(e))
    self.__base_class__ = [*bases, object][0]
    self.__base_classes__ = [*bases, ]
    self.__key_args__ = kwargs or {}

  def getMetaclass(self, ) -> type:
    """Returns the metaclass."""
    return self.__meta_class__

  def getClassName(self, ) -> str:
    """Returns the name of the class."""
    return self.__class_name__

  def getBaseClass(self, ) -> type:
    """Returns the base class."""
    return self.__base_class__

  def getKwargs(self, ) -> dict:
    """Returns the keyword arguments passed to the class."""
    return {**self.__key_args__, **dict()}

  def getMRO(self, ) -> Types:
    """Returns the method resolution order of the class."""
    mcls = self.__meta_class__
    name = '_THIS_IS_WHY_WE_CANT_HAVE_NICE_THINGS'
    Space = type(self)
    space = Space(mcls, name, self.getClassBases(), )
    bases = self.getClassBases()
    tempClass = mcls(name, (*bases,), space)
    return tempClass.__mro__

  def getMROSpace(self, ) -> Spaces:
    """Returns the namespace objects of the baseclasses. """
    cls = type(self)
    mcls = self.getMetaclass()
    out = []
    for base in self.getMRO():
      if isinstance(base, mcls):
        out.append(getattr(base, '__namespace__', ))
    return (*out,)

  def getPrimeBase(self, ) -> type:
    """Returns the prime baseclass of the class under creation. """
    baseMRO = [*self.getMRO(), ]
    mcls = self.getMetaclass()
    while baseMRO:
      base = baseMRO.pop()
      if base is object:
        continue
      if isinstance(base, mcls):
        return base
    raise MissingVariable('base', mcls)

  def getPrimeSpace(self, ) -> Self:
    """Getter-function for the namespace object of the prime baseclass of
    the class under creation. The prime baseclass is the baseclass whose
    baseclasses are not derived from this metaclass. """
    try:
      primeBase = self.getPrimeBase()
    except MissingVariable as missingVariable:
      return self
    cls = type(self)
    primeSpace = getattr(primeBase, '__namespace__', None)
    if isinstance(primeSpace, cls):
      return primeSpace
    if primeSpace is None:
      return self
    raise TypeError(typeMsg('primeSpace', primeSpace, cls))

  def isPrime(self, ) -> bool:
    """Returns True if the class is the prime baseclass of the class
    under creation. """
    return True if self.getPrimeSpace() is self else False

  def getParent(self, ) -> type:
    """Getter-function for the parent baseclass of the class under
    creation. """
    if self.isPrime():
      raise MissingVariable('parent', type(self))
    return self.getBaseClass()

  def getParentSpace(self, ) -> Self:
    """Getter-function for the namespace object of the parent baseclass of
    the class under creation. """
    cls = type(self)
    try:
      parentBase = self.getParent()
    except MissingVariable as missingVariable:
      return self
    parentSpace = getattr(parentBase, '__namespace__', None)
    if parentSpace is None:
      raise MissingVariable('__namespace__', cls)
    if isinstance(parentSpace, cls):
      return parentSpace
    raise TypeError(typeMsg('parentSpace', parentSpace, cls))

  def __getitem__(self, key: str, **kwargs) -> Any:
    """Returns the value of the key."""
    try:
      val = HistDict.__getitem__(self, key)
    except KeyError as keyError:
      val = keyError
    for hook in self.getHooks():
      if not isinstance(hook, AbstractHook):
        raise TypeError(typeMsg('hook', hook, AbstractHook))
      try:
        hook.getItemHook(key, val)
      except Exception as exception:
        print('hook exception: %s' % type(hook).__name__)
        print('  key: %s' % key)
        print('  value: %s' % val)
        print('  %s!\n%s' % (type(exception).__name__, str(exception)))
        raise HookException(exception, self, key, val, hook)
    if isinstance(val, KeyError):
      raise val
    return val

  def __setitem__(self, key: str, val: object, **kwargs) -> None:
    """Sets the value of the key."""
    try:
      oldVal = HistDict.__getitem__(self, key)
    except KeyError:
      oldVal = None
    for hook in self.getHooks():
      if TYPE_CHECKING:
        assert isinstance(hook, AbstractHook)
      if hook.setItemHook(key, val, oldVal):
        break
    else:
      HistDict.__setitem__(self, key, val)

  def preCompile(self, ) -> dict:
    """The return value from this method is passed to the compile method.
    Subclasses can implement this method to provide special objects at
    particular names in the namespace. By default, an empty dictionary is
    returned. """
    namespace = dict()
    for hook in self.getHooks():
      if TYPE_CHECKING:
        assert isinstance(hook, AbstractHook)
      namespace = hook.preCompileHook(namespace)
    return namespace

  def compile(self, ) -> dict:
    """This method is responsible for building the final namespace object.
    Subclasses may reimplement preCompile or postCompile as needed,
    but must not reimplement this method."""
    if self.__class_name__ == '_THIS_IS_WHY_WE_CANT_HAVE_NICE_THINGS':
      return dict()
    namespace = self.preCompile()
    for (key, val) in dict.items(self, ):
      namespace[key] = val
    namespace['__metaclass__'] = self.getMetaclass()
    namespace['__namespace__'] = self
    return self.postCompile(namespace)

  def postCompile(self, namespace: dict) -> dict:
    """The object returned from this method is passed to the __new__
    method in the owning metaclass. By default, this method returns dict
    object created by the compile method after performing certain
    validations. Subclasses can implement this method to provide further
    processing of the compiled object. """
    for hook in self.getHooks():
      if TYPE_CHECKING:
        assert isinstance(hook, AbstractHook)
      namespace = hook.postCompileHook(namespace)
    if '__metaclass__' not in namespace:
      raise MissingVariable('__metaclass__', self.getMetaclass())
    if '__namespace__' not in namespace:
      raise MissingVariable('__namespace__', type(self))
    return namespace

  def __str__(self, ) -> str:
    """Returns the string representation of the namespace object."""
    spaceName = type(self).__name__
    clsName = self.getClassName()
    baseName = self.getBaseClass().__name__
    mclsName = self.getMetaclass().__name__
    info = """Namespace object of type: '%s' created by the '__prepare__' 
    method on metaclass: '%s' with base: %s to create class: '%s'."""
    return monoSpace(info % (spaceName, mclsName, baseName, clsName))

  def __repr__(self, ) -> str:
    """Returns the string representation of the namespace object."""
    spaceName = type(self).__name__
    clsName = self.getClassName()
    mclsName = self.getMetaclass().__name__
    baseName = self.getBaseClass().__name__
    mclsName = self.getMetaclass().__name__
    args = """%s, %s, (%s,)""" % (mclsName, clsName, baseName)
    kwargs = [(k, v) for (k, v) in self.getKwargs().items()]
    kwargStr = ', '.join(['%s=%s' % (k, str(v)) for (k, v) in kwargs])
    if kwargStr:
      kwargStr = ', %s' % kwargStr
    return """%s(%s%s)""" % (spaceName, args, kwargStr)



"""AbstractDescriptor provides a common abstract baseclass for all
descriptor classes. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from abc import abstractmethod

from ..static import AbstractObject
from ..waitaminute import MissingVariable

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any


class _MetaDescriptor(type):
  """MetaDescriptor is a metaclass for the AbstractDescriptor class."""

  def __getitem__(cls, fieldType: type) -> Any:
    """Get the field type."""
    self = cls()
    self._addFieldType(fieldType)
    return self


class AbstractDescriptor(AbstractObject):
  """AbstractDescriptor provides a common abstract baseclass for all
  descriptor classes. """
  pass



"""ReservedNameHook protects reserved names. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ...waitaminute import ReservedName

from . import AbstractHook, ReservedNames

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any


class ReservedNameHook(AbstractHook):
  """ReservedNameHook protects reserved names."""

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Public variables
  reservedNames = ReservedNames()

  def setItemHook(self, key: str, value: Any, oldValue: Any) -> bool:
    """The setItemHook method is called when an item is set in the
    namespace."""
    if key in self.reservedNames:
      if key in self.space:
        raise ReservedName(key)
    return False
